@@ -91,6 +91,15 @@ void MVM_sc_add_all_scs_entry(MVMThreadContext *tc, MVMSerializationContextBody
     tc->instance->all_scs_next_idx++;
 }
 
+void MVM_sc_all_scs_destroy(MVMThreadContext *tc) {
+    MVM_fixed_size_free(
+        tc,
+        tc->instance->fsa,
+        tc->instance->all_scs_alloc * sizeof(MVMSerializationContextBody *),
+        tc->instance->all_scs
+    );
+}
+
 /* Given an SC, returns its unique handle. */
 MVMString * MVM_sc_get_handle(MVMThreadContext *tc, MVMSerializationContext *sc) {
     return sc->body->handle;
@@ -306,10 +315,8 @@ void MVM_sc_set_stable(MVMThreadContext *tc, MVMSerializationContext *sc, MVMint
             sc->body->alloc_stables += 32;
             if (sc->body->alloc_stables < (MVMuint64)idx + 1)
                 sc->body->alloc_stables = idx + 1;
-            sc->body->root_stables = MVM_realloc(sc->body->root_stables,
+            sc->body->root_stables = MVM_recalloc(sc->body->root_stables, orig_size,
                 sc->body->alloc_stables * sizeof(MVMSTable *));
-            memset(sc->body->root_stables + orig_size, 0,
-                (sc->body->alloc_stables - orig_size) * sizeof(MVMSTable *));
         }
         MVM_ASSIGN_REF(tc, &(sc->common.header), sc->body->root_stables[idx], st);
         sc->body->num_stables = idx + 1;
