@@ -1,11 +1,6 @@
 #include "moar.h"
 
 #define FIXKEY_INITIAL_SIZE_LOG2 3
-#define FIXKEY_INITIAL_KEY_RIGHT_SHIFT (8 * sizeof(MVMuint64) - 3)
-
-MVM_STATIC_INLINE MVMuint32 calc_entries_in_use(const struct MVMFixKeyHashTableControl *control) {
-    return MVM_fixkey_hash_official_size(control) + control->max_probe_distance - 1;
-}
 
 void hash_demolish_internal(MVMThreadContext *tc,
                             struct MVMFixKeyHashTableControl *control) {
@@ -25,7 +20,7 @@ void MVM_fixkey_hash_demolish(MVMThreadContext *tc, MVMFixKeyHashTable *hashtabl
     if (!control)
         return;
 
-    MVMuint32 entries_in_use = calc_entries_in_use(control);
+    MVMuint32 entries_in_use = MVM_fixkey_hash_kompromat(control);
     MVMuint8 *entry_raw = MVM_fixkey_hash_entries(control);
     MVMuint8 *metadata = MVM_fixkey_hash_metadata(control);
     MVMuint32 bucket = 0;
@@ -47,7 +42,6 @@ void MVM_fixkey_hash_demolish(MVMThreadContext *tc, MVMFixKeyHashTable *hashtabl
 
 MVM_STATIC_INLINE struct MVMFixKeyHashTableControl *hash_allocate_common(MVMThreadContext *tc,
                                                                          MVMuint16 entry_size,
-                                                                         MVMuint8 key_right_shift,
                                                                          MVMuint8 official_size_log2) {
     MVMuint32 official_size = 1 << (MVMuint32)official_size_log2;
     MVMuint32 max_items = official_size * MVM_FIXKEY_HASH_LOAD_FACTOR;
@@ -75,7 +69,8 @@ MVM_STATIC_INLINE struct MVMFixKeyHashTableControl *hash_allocate_common(MVMThre
     MVMuint8 initial_probe_distance = (1 << (8 - MVM_HASH_INITIAL_BITS_IN_METADATA)) - 1;
     control->max_probe_distance = max_probe_distance_limit > initial_probe_distance ? initial_probe_distance : max_probe_distance_limit;
     control->max_probe_distance_limit = max_probe_distance_limit;
-    control->key_right_shift = key_right_shift;
+    MVMuint8 bucket_right_shift = 8 * sizeof(MVMuint64) - official_size_log2;
+    control->key_right_shift = bucket_right_shift - control->metadata_hash_bits;
     control->entry_size = entry_size;
 
     MVMuint8 *metadata = (MVMuint8 *)(control + 1);
@@ -90,7 +85,6 @@ void MVM_fixkey_hash_build(MVMThreadContext *tc, MVMFixKeyHashTable *hashtable,
     }
     hashtable->table = hash_allocate_common(tc,
                                             entry_size,
-                                            FIXKEY_INITIAL_KEY_RIGHT_SHIFT,
                                             FIXKEY_INITIAL_SIZE_LOG2);
 }
 
@@ -245,6 +239,7 @@ static struct MVMFixKeyHashTableControl *maybe_grow_hash(MVMThreadContext *tc,
         } while (--loop_count);
         assert(control->metadata_hash_bits);
         --control->metadata_hash_bits;
+        ++control->key_right_shift;
 
         control->max_probe_distance = new_probe_distance;
         /* Reset this to its proper value. */
@@ -253,7 +248,7 @@ static struct MVMFixKeyHashTableControl *maybe_grow_hash(MVMThreadContext *tc,
         return NULL;
     }
 
-    MVMuint32 entries_in_use = calc_entries_in_use(control);
+    MVMuint32 entries_in_use = MVM_fixkey_hash_kompromat(control);
     MVMuint8 *entry_raw_orig = MVM_fixkey_hash_entries(control);
     MVMuint8 *metadata_orig = MVM_fixkey_hash_metadata(control);
 
@@ -261,7 +256,6 @@ static struct MVMFixKeyHashTableControl *maybe_grow_hash(MVMThreadContext *tc,
 
     control = hash_allocate_common(tc,
                                    control_orig->entry_size,
-                                   control_orig->key_right_shift - 1,
                                    control_orig->official_size_log2 + 1);
 
     MVMuint8 *entry_raw = entry_raw_orig;
@@ -375,11 +369,14 @@ MVMuint64 MVM_fixkey_hash_fsck(MVMThreadContext *tc, MVMFixKeyHashTable *hashtab
         return 0;
     }
 
-    MVMuint32 entries_in_use = calc_entries_in_use(control);
+    MVMuint32 entries_in_use = MVM_fixkey_hash_kompromat(control);
     MVMuint8 *entry_raw = MVM_fixkey_hash_entries(control);
     MVMuint8 *metadata = MVM_fixkey_hash_metadata(control);
     MVMuint32 bucket = 0;
     MVMint64 prev_offset = 0;
+    MVMuint8 bucket_right_shift
+        = control->key_right_shift + control->metadata_hash_bits;
+
     while (bucket < entries_in_use) {
         if (!*metadata) {
             /* empty slot. */
@@ -401,7 +398,7 @@ MVMuint64 MVM_fixkey_hash_fsck(MVMThreadContext *tc, MVMFixKeyHashTable *hashtab
                     fprintf(stderr, "%s%3X!!\n", prefix_hashes, bucket);
                 } else {
                     MVMuint64 hash_val = MVM_string_hash_code(tc, *entry);
-                    MVMuint32 ideal_bucket = hash_val >> control->key_right_shift;
+                    MVMuint32 ideal_bucket = hash_val >> bucket_right_shift;
                     MVMint64 offset = 1 + bucket - ideal_bucket;
                     int wrong_bucket = offset != *metadata;
                     int wrong_order = offset < 1 || offset > prev_offset + 1;
@@ -422,7 +419,7 @@ MVMuint64 MVM_fixkey_hash_fsck(MVMThreadContext *tc, MVMFixKeyHashTable *hashtab
         ++metadata;
         entry_raw -= sizeof(MVMString ***);
     }
-    if (*metadata != 1) {
+    if (*metadata != 0) {
         ++errors;
         if (display) {
             fprintf(stderr, "%s    %02x!\n", prefix_hashes, *metadata);
