@@ -102,7 +102,7 @@ struct MVMEventSubscriptions {
     MVMObject *SpeshOverviewEvent;
 
     MVMuint64 vm_startup_hrtime;
-    MVMnum64  vm_startup_now;
+    MVMuint64 vm_startup_now;
 };
 
 struct MVMSerializationContextWeakHashEntry {
@@ -211,6 +211,9 @@ struct MVMInstance {
     MVM_VECTOR_DECL(void *, free_at_safepoint);
     uv_mutex_t mutex_free_at_safepoint;
 
+    /* Whether the --full-cleanup flag was passed. */
+    MVMuint32 full_cleanup;
+
     /************************************************************************
      * Object system
      ************************************************************************/
@@ -405,6 +408,12 @@ struct MVMInstance {
     /* Normal Form Grapheme state (synthetics table, lookup, etc.). */
     MVMNFGState *nfg;
 
+    /* Unicode hashes. */
+    MVMUniHashTable property_codes_by_names_aliases;
+    MVMUniHashTable property_codes_by_seq_names;
+    MVMUniHashTable codepoints_by_name;
+    MVMUniHashTable *unicode_property_values_hashes;
+
     /************************************************************************
      * Type objects for built-in types and special values
      ************************************************************************/
@@ -438,6 +447,8 @@ struct MVMInstance {
 
     MVMObject *SpeshPluginState;
 
+    MVMObject *SpeshCandidate;
+
     /* Set of bootstrapping types. */
     MVMBootTypes boot_types;
 
