@@ -230,7 +230,7 @@ static void callback_handler(ffi_cif *cif, void *cb_result, void **cb_args, void
 
     /* Build a callsite and arguments buffer. */
     args = MVM_malloc(data->num_types * sizeof(MVMRegister));
-    num_roots = 0;
+    num_roots = 1; /* res.o is always in roots */
     for (i = 1; i < data->num_types; i++) {
         MVMObject *type     = data->types[i];
         MVMint16   typeinfo = data->typeinfos[i];
@@ -321,6 +321,7 @@ static void callback_handler(ffi_cif *cif, void *cb_result, void **cb_args, void
 
     /* Call into a nested interpreter (since we already are in one). Need to
      * save a bunch of state around each side of this. */
+    MVM_gc_root_temp_push(tc, (MVMCollectable **)&(res.o));
     cid.invokee = data->target;
     cid.args    = args;
     cid.cs      = data->cs;
