@@ -40,7 +40,7 @@ MVM_STATIC_INLINE void hash_demolish_internal(MVMThreadContext *tc,
 
     size_t total_size
         = entries_size + sizeof(struct MVMStrHashTableControl) + metadata_size;
-    MVM_fixed_size_free(tc, tc->instance->fsa, total_size, start);
+    MVM_fixed_size_free_at_safepoint(tc, tc->instance->fsa, total_size, start);
 }
 
 /* Frees the entire contents of the hash, leaving you just the hashtable itself,
@@ -49,6 +49,10 @@ void MVM_str_hash_demolish(MVMThreadContext *tc, MVMStrHashTable *hashtable) {
     struct MVMStrHashTableControl *control = hashtable->table;
     if (!control)
         return;
+    if (MVM_UNLIKELY(control->stale)) {
+        MVM_oops(tc, "MVM_str_hash_demolish called with a stale hashtable pointer");
+    }
+
     hash_demolish_internal(tc, control);
     hashtable->table = NULL;
 }
@@ -110,7 +114,6 @@ void MVM_str_hash_demolish(MVMThreadContext *tc, MVMStrHashTable *hashtable) {
 
 MVM_STATIC_INLINE struct MVMStrHashTableControl *hash_allocate_common(MVMThreadContext *tc,
                                                                       MVMuint8 entry_size,
-                                                                      MVMuint8 key_right_shift,
                                                                       MVMuint8 official_size_log2) {
     MVMuint32 official_size = 1 << (MVMuint32)official_size_log2;
     MVMuint32 max_items = official_size * MVM_STR_HASH_LOAD_FACTOR;
@@ -139,8 +142,10 @@ MVM_STATIC_INLINE struct MVMStrHashTableControl *hash_allocate_common(MVMThreadC
     MVMuint8 initial_probe_distance = (1 << (8 - MVM_HASH_INITIAL_BITS_IN_METADATA)) - 1;
     control->max_probe_distance = max_probe_distance_limit > initial_probe_distance ? initial_probe_distance : max_probe_distance_limit;
     control->max_probe_distance_limit = max_probe_distance_limit;
-    control->key_right_shift = key_right_shift;
+    MVMuint8 bucket_right_shift = 8 * sizeof(MVMuint64) - official_size_log2;
+    control->key_right_shift = bucket_right_shift - control->metadata_hash_bits;
     control->entry_size = entry_size;
+    control->stale = 0;
 
     MVMuint8 *metadata = (MVMuint8 *)(control + 1);
     memset(metadata, 0, metadata_size);
@@ -182,10 +187,7 @@ void MVM_str_hash_build(MVMThreadContext *tc,
             initial_size_base2 = STR_MIN_SIZE_BASE_2;
         }
 
-        control = hash_allocate_common(tc,
-                                       entry_size,
-                                       (8 * sizeof(MVMuint64) - initial_size_base2),
-                                       initial_size_base2);
+        control = hash_allocate_common(tc, entry_size, initial_size_base2);
     }
 
 #if HASH_DEBUG_ITER
@@ -323,15 +325,24 @@ MVM_STATIC_INLINE struct MVMStrHashHandle *hash_insert_internal(MVMThreadContext
 static struct MVMStrHashTableControl *maybe_grow_hash(MVMThreadContext *tc,
                                                       struct MVMStrHashTableControl *control) {
     if (MVM_UNLIKELY(control->cur_items == 0 && control->max_items == 0)) {
+        /* This is the case where we had initially allocated just a control
+         * structure, and now we find that we need to allocate some memory for
+         * the first entry. Which implies that we've had no insertions or
+         * deletions prior to this, hence our debugging state must be (0, 0) */
         struct MVMStrHashTableControl *control_orig = control;
+
+        control_orig->stale = 1;
         control = hash_allocate_common(tc,
                                        control_orig->entry_size,
-                                       (8 * sizeof(MVMuint64) - STR_MIN_SIZE_BASE_2),
                                        STR_MIN_SIZE_BASE_2);
 #if HASH_DEBUG_ITER
         control->ht_id = control_orig->ht_id;
+        assert(control_orig->serial == 0);
+        assert(control_orig->last_delete_at == 0);
+        control->serial = 0;
+        control->last_delete_at = 0;
 #endif
-        MVM_fixed_size_free(tc, tc->instance->fsa, sizeof(*control_orig), control_orig);
+        MVM_fixed_size_free_at_safepoint(tc, tc->instance->fsa, sizeof(*control_orig), control_orig);
         return control;
     }
 
@@ -372,6 +383,7 @@ static struct MVMStrHashTableControl *maybe_grow_hash(MVMThreadContext *tc,
         } while (--loop_count);
         assert(control->metadata_hash_bits);
         --control->metadata_hash_bits;
+        ++control->key_right_shift;
 
         control->max_probe_distance = new_probe_distance;
         /* Reset this to its proper value. */
@@ -387,9 +399,9 @@ static struct MVMStrHashTableControl *maybe_grow_hash(MVMThreadContext *tc,
 
     struct MVMStrHashTableControl *control_orig = control;
 
+    control_orig->stale = 1;
     control = hash_allocate_common(tc,
                                    entry_size,
-                                   control_orig->key_right_shift - 1,
                                    control_orig->official_size_log2 + 1);
 
 
@@ -450,6 +462,9 @@ void *MVM_str_hash_lvalue_fetch_nocheck(MVMThreadContext *tc,
     if (MVM_UNLIKELY(!control)) {
         MVM_oops(tc, "Attempting insert on MVM_str_hash without first calling MVM_str_hash_build");
     }
+    else if (MVM_UNLIKELY(control->stale)) {
+        MVM_oops(tc, "MVM_str_hash_lvalue_fetch_nocheck called with a stale hashtable pointer");
+    }
     else if (MVM_UNLIKELY(control->cur_items >= control->max_items)) {
         /* We should avoid growing the hash if we don't need to.
          * It's expensive, and for hashes with iterators, growing the hash
@@ -457,20 +472,32 @@ void *MVM_str_hash_lvalue_fetch_nocheck(MVMThreadContext *tc,
          * need to create a key. */
         struct MVMStrHashHandle *entry = MVM_str_hash_fetch_nocheck(tc, hashtable, key);
         if (entry) {
+            if (MVM_UNLIKELY(control->stale)) {
+                MVM_oops(tc, "MVM_str_hash_lvalue_fetch_nocheck called with a hashtable pointer that turned stale");
+            }
             return entry;
         }
 
         struct MVMStrHashTableControl *new_control = maybe_grow_hash(tc, control);
         if (new_control) {
-            /* We could unconditionally assign this, but that would mean CPU
-             * cache writes even when it was unchanged, and the address of
-             * hashtable will not be in the same cache lines as we are writing
-             * for the hash internals, so it will churn the write cache. */
-            hashtable->table = control = new_control;
+            if (!MVM_trycas(&(hashtable->table), control, new_control)) {
+                /* Oh erk, a second thread has just executed this code on this
+                 * hashtable, allocated a new control structure, and both it and
+                 * us have passed the old control structure back to the FSA.
+                 * So we have a double free pending at the next safe point, and
+                 * who knows what else is racing uncontrolled on this hash.
+                 * Bad programmer, no cookie. */
+                MVM_oops(tc, "MVM_str_hash_lvalue_fetch_nocheck called concurrently on the same hash");
+            }
+            control = new_control;
         }
     }
 
-    return hash_insert_internal(tc, control, key);
+    void *result = hash_insert_internal(tc, control, key);
+    if (MVM_UNLIKELY(control->stale)) {
+        MVM_oops(tc, "MVM_str_hash_lvalue_fetch_nocheck called with a hashtable pointer that turned stale");
+    }
+    return result;
 }
 
 /* UNCONDITIONALLY creates a new hash entry with the given key and value.
@@ -497,11 +524,15 @@ void *MVM_str_hash_insert_nocheck(MVMThreadContext *tc,
 void MVM_str_hash_delete_nocheck(MVMThreadContext *tc,
                                  MVMStrHashTable *hashtable,
                                  MVMString *key) {
+    struct MVMStrHashTableControl *control = hashtable->table;
+
+    if (MVM_UNLIKELY(control && control->stale)) {
+        MVM_oops(tc, "MVM_str_hash_delete_nocheck called with a stale hashtable pointer");
+    }
     if (MVM_str_hash_is_empty(tc, hashtable)) {
         return;
     }
 
-    struct MVMStrHashTableControl *control = hashtable->table;
     struct MVM_hash_loop_state ls = MVM_str_hash_create_loop_state(tc, control, key);
 
     while (1) {
@@ -550,6 +581,83 @@ void MVM_str_hash_delete_nocheck(MVMThreadContext *tc,
                 /* and this slot is now emtpy. */
                 *metadata_target = 0;
                 --control->cur_items;
+                if (control->max_items == 0
+                    && control->cur_items < control->max_probe_distance) {
+                    /* So this is a fun corner case...
+                     *
+                     * Rakudo code often creates empty hashes, so we have a
+                     * space optimisation to (not) allocate "8" (ie 13) slots
+                     * initially. Instead we only allocate a control structure.
+                     * However most of the metadata in that is stored log base 2
+                     * (another size optimisation) hence *it* can't store zero.
+                     * So we mark this case by setting both
+                     * control->cur_items == 0 && control->max_items == 0
+                     * `max_items` zero triggers immediate allocation on any
+                     * insert ("no questions asked") and `cur_items` zero is the
+                     * true state. See commit 7d6107e65ddedece
+                     *
+                     * The assumption of that commit was that it was impossible
+                     * for the control structure to be in that (zero,zero) state
+                     * as soon as any insert happened, because after that
+                     * max_items would only be set to zero if an insert flagged
+                     * up that the `max_probe_distance` might overflow on the
+                     * *next* insert (ie trigger that immediate allocation),
+                     * and if there are items allocated, then (obviously)
+                     * `cur_items` isn't zero.
+                     *
+                     * What I missed was that a sequence of inserts followed by
+                     * a sequence of deletes can reach (zero,zero).
+                     * If the *last* insert hits the `max_probe_distance` limit,
+                     * then it sets `max_items` to zero so that the next insert
+                     * will allocate.
+                     *
+                     * But what if there is no next insert?
+                     *
+                     * Then nothing resets the `max_items`, and it remains as
+                     * zero.
+                     *
+                     * And if the only write actions that happen on the hash
+                     * after this are to delete each entry in turn, then
+                     * eventually `cur_items` reaches zero too (accurately), and
+                     * then the "special state" flags are accidentally recreated
+                     * but not with the "special state" memory layout.
+                     *
+                     * (And with enough debugging enabled the assignment to
+                     *  `control->last_delete_at` just below fails an assertion
+                     *  in `MVM_str_hash_metadata`)
+                     *
+                     * So clearly we need to unwind the "immediate allocation"
+                     * flag.
+                     *
+                     * We certainly can't do it on *any* delete
+                     *
+                     * We can't actually do it on *any* delete that drops a
+                     * probe distance below the limit because (this one is
+                     * subtle) worst case it's possible for
+                     * a) a hash to be in the state where several entries have
+                     *    probe distances one below the threshold
+                     * b) a *single* insert causes 2+ of these to move up
+                     *    (the "make room" code in insert)
+                     *    (or 1 to move up, *and* the new insert to be at the
+                     *     the trigger distance)
+                     *    which sets the flag
+                     * c) but the immediately subsequent deletion causes only
+                     *    *one* of these to drop back below the max.
+                     *
+                     * The really conservative approach would only be to reset
+                     * if the hash is about to hit zero items.
+                     *
+                     * But I believe that the earliest we can be *sure* that no
+                     * chain is longer than the limit is when the *total*
+                     * entries in the hash are less than that limit.  Because
+                     * (of course), the worst case is that they are all in a row
+                     * contesting the same ideal bucket, and the highest probe
+                     * distance has to be less than the limit.
+                     *
+                     * So here we are: */
+                    MVMuint32 official_size = 1 << (MVMuint32)control->official_size_log2;
+                    control->max_items = official_size * MVM_STR_HASH_LOAD_FACTOR;
+                }
 
 #if HASH_DEBUG_ITER
                 ++control->serial;
@@ -560,6 +668,10 @@ void MVM_str_hash_delete_nocheck(MVMThreadContext *tc,
                 control->last_delete_at = 1 + ls.metadata - MVM_str_hash_metadata(control);
 #endif
 
+                if (MVM_UNLIKELY(control->stale)) {
+                    MVM_oops(tc, "MVM_str_hash_delete_nocheck called with a hashtable pointer that turned stale");
+                }
+
                 /* Job's a good 'un. */
                 return;
             }
@@ -572,6 +684,11 @@ void MVM_str_hash_delete_nocheck(MVMThreadContext *tc,
                the hash table - it would have stolen this slot, and the key we
                find here now would have been displaced further on. Hence, the key
                we seek can't be in the hash table. */
+
+            if (MVM_UNLIKELY(control->stale)) {
+                MVM_oops(tc, "MVM_str_hash_delete_nocheck called with a hashtable pointer that turned stale");
+            }
+
             return;
         }
         ls.probe_distance += ls.metadata_increment;
@@ -590,9 +707,10 @@ static MVMuint64 hash_fsck_internal(MVMThreadContext *tc, struct MVMStrHashTable
     MVMuint64 errors = 0;
     MVMuint64 seen = 0;
 
-    if (MVM_str_hash_entries(control) == NULL) {
+    if (!control || (control->cur_items == 0 && control->max_items == 0)) {
         if (display) {
-            fprintf(stderr, "%s NULL %p (empty)\n", prefix_hashes, control);
+            fprintf(stderr, "%s %p (empty%s)\n", prefix_hashes, control,
+                    control ? " optimisation" : "");
         }
         return 0;
     }
@@ -603,6 +721,9 @@ static MVMuint64 hash_fsck_internal(MVMThreadContext *tc, struct MVMStrHashTable
     MVMuint8 *metadata = MVM_str_hash_metadata(control);
     MVMuint32 bucket = 0;
     MVMint64 prev_offset = 0;
+    MVMuint8 bucket_right_shift
+        = control->key_right_shift + control->metadata_hash_bits;
+
     while (bucket < allocated_items) {
         if (!*metadata) {
             /* empty slot. */
@@ -667,7 +788,7 @@ static MVMuint64 hash_fsck_internal(MVMThreadContext *tc, struct MVMStrHashTable
             } else {
                 /* OK, it is a concrete string (still). */
                 MVMuint64 hash_val = MVM_str_hash_code(tc, control->salt, key);
-                MVMuint32 ideal_bucket = hash_val >> control->key_right_shift;
+                MVMuint32 ideal_bucket = hash_val >> bucket_right_shift;
                 MVMint64 offset = 1 + bucket - ideal_bucket;
                 MVMuint32 actual_bucket = *metadata >> metadata_hash_bits;
                 char wrong_bucket = offset == actual_bucket ? ' ' : '!';
@@ -675,10 +796,8 @@ static MVMuint64 hash_fsck_internal(MVMThreadContext *tc, struct MVMStrHashTable
                 if (offset < 1) {
                     wrong_order = '<';
                 } else if (offset > control->max_probe_distance) {
-                    ++errors;
                     wrong_order = '>';
                 } else if (offset > prev_offset + 1) {
-                    ++errors;
                     wrong_order = '!';
                 } else {
                     wrong_order = ' ';
@@ -738,7 +857,7 @@ static MVMuint64 hash_fsck_internal(MVMThreadContext *tc, struct MVMStrHashTable
         ++metadata;
         entry_raw -= control->entry_size;
     }
-    if (*metadata != 1) {
+    if (*metadata != 0) {
         ++errors;
         if (display) {
             fprintf(stderr, "%s    %02x!\n", prefix_hashes, *metadata);
