@@ -153,6 +153,7 @@ static MVMint32 already_succs(MVMThreadContext *tc, MVMSpeshBB *bb, MVMSpeshBB *
 #define MVM_CFG_BB_JUMPLIST 4
 static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf,
                       MVMint32 *existing_deopts, MVMint32 num_existing_deopts,
+                      MVMint32 *existing_deopt_synths, MVMint32 num_existing_deopt_synths,
                       MVMint32 *deopt_usage_info, MVMSpeshIns ***deopt_usage_ins_out) {
     MVMSpeshBB  *cur_bb, *prev_bb;
     MVMSpeshIns *last_ins;
@@ -600,10 +601,8 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
                 while (ann) {
                     switch (ann->type) {
                         case MVM_SPESH_ANN_FH_START:
-                            if (!is_catch_handler(tc, g, ann->data.frame_handler_index)) {
-                                active_handlers[ann->data.frame_handler_index] = 1;
-                                num_active_handlers++;
-                            }
+                            active_handlers[ann->data.frame_handler_index] = 1;
+                            num_active_handlers++;
                             break;
                         case MVM_SPESH_ANN_FH_END:
                             has_end = 1;
@@ -616,10 +615,8 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
                     while (ann) {
                         switch (ann->type) {
                             case MVM_SPESH_ANN_FH_END:
-                                if (!is_catch_handler(tc, g, ann->data.frame_handler_index)) {
-                                    active_handlers[ann->data.frame_handler_index] = 0;
-                                    num_active_handlers--;
-                                }
+                                active_handlers[ann->data.frame_handler_index] = 0;
+                                num_active_handlers--;
                                 break;
                         }
                         ann = ann->next;
@@ -731,6 +728,21 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
             }
         }
     }
+    if (existing_deopt_synths) {
+        for (i = 0; i < num_existing_deopt_synths; i ++) {
+            if (existing_deopt_synths[2 * i + 1] >= 0) {
+                MVMSpeshIns *post_ins     = ins_flat[byte_to_ins_flags[existing_deopt_synths[2 * i + 1]] >> 3];
+                MVMSpeshIns *deopt_ins    = post_ins->prev ? post_ins->prev :
+                    MVM_spesh_graph_linear_prev(tc, g,
+                        ins_to_bb[byte_to_ins_flags[existing_deopt_synths[2 * i + 1]] >> 3])->last_ins;
+                MVMSpeshAnn *deopt_ann    = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshAnn));
+                deopt_ann->next           = deopt_ins->annotations;
+                deopt_ann->type           = MVM_SPESH_ANN_DEOPT_SYNTH;
+                deopt_ann->data.deopt_idx = existing_deopt_synths[2 * i];
+                deopt_ins->annotations    = deopt_ann;
+            }
+        }
+    }
 
     /* We may also need to reconstruct deopt usage info, to allow optimizing of
      * inlinees. That is persisted using bytecode offsets, and this is the last
@@ -766,10 +778,14 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
 static MVMint32 is_handler_reg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMuint16 reg) {
     MVMuint32 num_handlers = g->num_handlers;
     MVMuint32 i;
-    for (i = 0; i < num_handlers; i++)
+    for (i = 0; i < num_handlers; i++) {
         if (g->handlers[i].action == MVM_EX_ACTION_INVOKE)
             if (g->handlers[i].block_reg == reg)
                 return 1;
+        if (g->handlers[i].category_mask & MVM_EX_CAT_LABELED)
+            if (g->handlers[i].label_reg == reg)
+                return 1;
+    }
     return 0;
 }
 static void insert_object_null_instructions(MVMThreadContext *tc, MVMSpeshGraph *g) {
@@ -1310,7 +1326,7 @@ MVMSpeshGraph * MVM_spesh_graph_create(MVMThreadContext *tc, MVMStaticFrame *sf,
     }
 
     /* Build the CFG out of the static frame, and transform it to SSA. */
-    build_cfg(tc, g, sf, NULL, 0, NULL, NULL);
+    build_cfg(tc, g, sf, NULL, 0, NULL, 0, NULL, NULL);
     if (insert_object_nulls)
         insert_object_null_instructions(tc, g);
     if (!cfg_only) {
@@ -1330,29 +1346,29 @@ MVMSpeshGraph * MVM_spesh_graph_create_from_cand(MVMThreadContext *tc, MVMStatic
     /* Create top-level graph object. */
     MVMSpeshGraph *g     = MVM_calloc(1, sizeof(MVMSpeshGraph));
     g->sf                = sf;
-    g->bytecode          = cand->bytecode;
-    g->bytecode_size     = cand->bytecode_size;
-    g->handlers          = cand->handlers;
-    g->num_handlers      = cand->num_handlers;
-    g->num_locals        = cand->num_locals;
-    g->num_lexicals      = cand->num_lexicals;
-    g->inlines           = cand->inlines;
-    g->num_inlines       = cand->num_inlines;
-    g->deopt_addrs       = cand->deopts;
-    g->num_deopt_addrs   = cand->num_deopts;
-    g->alloc_deopt_addrs = cand->num_deopts;
-    g->deopt_named_used_bit_field = cand->deopt_named_used_bit_field;
-    g->deopt_pea         = cand->deopt_pea;
-    g->local_types       = cand->local_types;
-    g->lexical_types     = cand->lexical_types;
-    g->num_spesh_slots   = cand->num_spesh_slots;
-    g->alloc_spesh_slots = cand->num_spesh_slots;
+    g->bytecode          = cand->body.bytecode;
+    g->bytecode_size     = cand->body.bytecode_size;
+    g->handlers          = cand->body.handlers;
+    g->num_handlers      = cand->body.num_handlers;
+    g->num_locals        = cand->body.num_locals;
+    g->num_lexicals      = cand->body.num_lexicals;
+    g->inlines           = cand->body.inlines;
+    g->num_inlines       = cand->body.num_inlines;
+    g->deopt_addrs       = cand->body.deopts;
+    g->num_deopt_addrs   = cand->body.num_deopts;
+    g->alloc_deopt_addrs = cand->body.num_deopts;
+    g->deopt_named_used_bit_field = cand->body.deopt_named_used_bit_field;
+    g->deopt_pea         = cand->body.deopt_pea;
+    g->local_types       = cand->body.local_types;
+    g->lexical_types     = cand->body.lexical_types;
+    g->num_spesh_slots   = cand->body.num_spesh_slots;
+    g->alloc_spesh_slots = cand->body.num_spesh_slots;
     g->phi_infos         = MVM_spesh_alloc(tc, g, MVMPhiNodeCacheSize * sizeof(MVMOpInfo));
     g->cand              = cand;
 
     g->spesh_slots       = MVM_malloc(g->alloc_spesh_slots * sizeof(MVMCollectable *));
 
-    memcpy(g->spesh_slots, cand->spesh_slots, sizeof(MVMCollectable *) * g->num_spesh_slots);
+    memcpy(g->spesh_slots, cand->body.spesh_slots, sizeof(MVMCollectable *) * g->num_spesh_slots);
 
     /* Ensure the frame is validated, since we'll rely on this. */
     if (sf->body.instrumentation_level == 0) {
@@ -1361,8 +1377,9 @@ MVMSpeshGraph * MVM_spesh_graph_create_from_cand(MVMThreadContext *tc, MVMStatic
     }
 
     /* Build the CFG out of the static frame, and transform it to SSA. */
-    build_cfg(tc, g, sf, cand->deopts, cand->num_deopts, cand->deopt_usage_info,
-            deopt_usage_ins_out);
+    build_cfg(tc, g, sf, cand->body.deopts, cand->body.num_deopts,
+            cand->body.deopt_synths, cand->body.num_deopt_synths,
+            cand->body.deopt_usage_info, deopt_usage_ins_out);
     if (!cfg_only) {
         MVM_spesh_eliminate_dead_bbs(tc, g, 0);
         add_predecessors(tc, g);
@@ -1433,6 +1450,10 @@ void MVM_spesh_graph_mark(MVMThreadContext *tc, MVMSpeshGraph *g, MVMGCWorklist
     /* Mark inlines. */
     for (i = 0; i < g->num_inlines; i++)
         MVM_gc_worklist_add(tc, worklist, &(g->inlines[i].sf));
+
+    /* Mark spesh candidate. */
+    if (g->cand)
+        MVM_gc_worklist_add(tc, worklist, &(g->cand));
 }
 
 /* Describes GCables for a heap snapshot. */
@@ -1478,21 +1499,23 @@ void MVM_spesh_graph_destroy(MVMThreadContext *tc, MVMSpeshGraph *g) {
     /* If there is a candidate that we either generated or that this graph was
      * generated from, it has ownership of the malloc'd memory. If not, then we
      * need to clean up */
-    if (g->spesh_slots && (!g->cand || g->cand->spesh_slots != g->spesh_slots))
+    if (g->spesh_slots && (!g->cand || g->cand->body.spesh_slots != g->spesh_slots))
         MVM_free(g->spesh_slots);
-    if (g->deopt_addrs && (!g->cand || g->cand->deopts != g->deopt_addrs))
+    if (g->deopt_addrs && (!g->cand || g->cand->body.deopts != g->deopt_addrs))
         MVM_free(g->deopt_addrs);
-    if (g->inlines && (!g->cand || g->cand->inlines != g->inlines))
+    if (g->inlines && (!g->cand || g->cand->body.inlines != g->inlines))
         MVM_free(g->inlines);
-    if (g->local_types &&  (!g->cand || g->cand->local_types != g->local_types))
+    if (g->local_types &&  (!g->cand || g->cand->body.local_types != g->local_types))
         MVM_free(g->local_types);
-    if (g->lexical_types &&  (!g->cand || g->cand->lexical_types != g->lexical_types))
+    if (g->lexical_types &&  (!g->cand || g->cand->body.lexical_types != g->lexical_types))
         MVM_free(g->lexical_types);
+    if (!g->cand)
+        MVM_spesh_pea_destroy_deopt_info(tc, &(g->deopt_pea));
 
     /* Handlers can come directly from static frame, from spesh candidate, and
      * from malloc/realloc. We only free it in the last case */
     if (g->handlers && g->handlers != g->sf->body.handlers &&
-        (!g->cand || g->cand->handlers != g->handlers))
+        (!g->cand || g->cand->body.handlers != g->handlers))
         MVM_free(g->handlers);
 
     /* Free the graph itself. */
