@@ -92,7 +92,7 @@ sub add_emoji_sequences {
     while (my $file = readdir $UNIDATA_DIR) {
         push @versions, $file if -d "$directory/$file" && $file =~ s/ ^ emoji- //x;
     }
-    croak "Couldn't find any emoji folders. Please run UCD-download.p6 again"
+    croak "Couldn't find any emoji folders. Please run UCD-download.raku again"
         if !@versions;
     my $highest_emoji_version = get_highest_version(\@versions);
     say "Highest emoji version found is $highest_emoji_version";
@@ -1151,14 +1151,6 @@ sub emit_names_hash_builder {
     {0x10FFFE,0}
 };
 
-/* Lazily constructed hashtable of Unicode names to codepoints.
-    Okay not to be threadsafe since its value is deterministic
-        and I don't care about the tiny potential for a memory leak
-        in the event of a race condition. */
-
- /* static, so will be 0 initialised. */
-static MVMUniHashTable codepoints_by_name;
-
 static void generate_codepoints_by_name(MVMThreadContext *tc) {
     MVMint32 extent_index = 0;
     MVMint32 codepoint = 0;
@@ -1183,7 +1175,7 @@ static void generate_codepoints_by_name(MVMThreadContext *tc) {
                      * <control> <CJK UNIFIED IDEOGRAPH> <CJK COMPATIBILITY IDEOGRAPH>
                      * <surrogate> <TANGUT IDEOGRAPH> <private-use> */
                     if (name && *name != '<') {
-                        MVM_uni_hash_insert(tc, &codepoints_by_name, name, codepoint);
+                        MVM_uni_hash_insert(tc, &tc->instance->codepoints_by_name, name, codepoint);
                     }
                     codepoint++;
                     codepoint_table_index++;
@@ -1200,7 +1192,7 @@ static void generate_codepoints_by_name(MVMThreadContext *tc) {
             case $FATE_SPAN: {
                 const char *name = codepoint_names[codepoint_table_index];
                 if (name && *name != '<') {
-                    MVM_uni_hash_insert(tc, &codepoints_by_name, name, codepoint);
+                    MVM_uni_hash_insert(tc, &tc->instance->codepoints_by_name, name, codepoint);
                 }
                 codepoint += length;
                 codepoint_table_index++;
@@ -1209,7 +1201,7 @@ static void generate_codepoints_by_name(MVMThreadContext *tc) {
         }
     }
     for (; i >= 0; i--) {
-        MVM_uni_hash_insert(tc, &codepoints_by_name, uni_namealias_pairs[i].name, uni_namealias_pairs[i].codepoint);
+        MVM_uni_hash_insert(tc, &tc->instance->codepoints_by_name, uni_namealias_pairs[i].name, uni_namealias_pairs[i].codepoint);
     }
 
 }
@@ -1367,7 +1359,7 @@ sub add_unicode_sequence {
         # uppercase.
         $name = uc $name;
         # Emoji sequences have commas in some and these cannot be included
-        # since they seperate seperate named items in ISO notation that P6 uses
+        # since they seperate seperate named items in ISO notation that Raku uses
         $name =~ s/,//xg;
         $named_sequences->{$name}->{'type'} = $type;
         # Only push if we haven't seen this already
@@ -1585,7 +1577,7 @@ sub emit_unicode_property_value_keypairs {
             $done{"$propname$_"} ||= push @lines, $lines{$propname}->{$_};
         }
     }
-    my $out = "\nstatic MVMUniHashTable *unicode_property_values_hashes;\n" .
+    my $out = "\n" .
     "static const MVMUnicodeNamedValue unicode_property_value_keypairs[" . scalar(@lines) . "] = {\n" .
     "    " . stack_lines(\@lines, ",", ",\n    ", 0, $WRAP_TO_COLUMNS) . "\n" .
     "};";
@@ -2167,13 +2159,13 @@ sub collation {
         $line_no++;
         my ($code, $temp);
         my $weights = {};
-        # implicit weights are handled in ./tools/Generate-Collation-Data.p6
+        # implicit weights are handled in ./tools/Generate-Collation-Data.raku
         return if $line =~ s/ ^ \@implicitweights \s+ //xms;
         return if $line =~ / ^ \s* [#@] /x or $line =~ / ^ \s* $ /x; # Blank/comment lines
         ($code, $temp) = split / [;#]+ /x, $line;
         $code = trim $code;
         my @codes = split ' ', $code;
-        # We support collation for multiple codepoints in ./tools/Generate-Collation-Data.p6
+        # We support collation for multiple codepoints in ./tools/Generate-Collation-Data.raku
         if (1 < @codes) {
             # For now set MVM_COLLATION_QC = 0 for these cp
             apply_to_range($codes[0], sub {
